###############################################################
###   Tcl Variables
###############################################################
#set tclParams [list <param1> <value> <param2> <value> ... <paramN> <value>]
set tclParams [list place.closeImportedSites  1 \
                    hd.StrictContainRouting   1 \
              ]

#Define location for "Tcl" directory. Defaults to "./Tcl"
set tclHome "F:/OOC/Tcl"
if {[file exists $tclHome]} {
   set tclDir $tclHome
} elseif {[file exists "./Tcl"]} {
   set tclDir  "./Tcl"
} else {
   error "ERROR: No valid location found for required Tcl scripts. Set \$tclDir in design.tcl to a valid location."
}

###############################################################
### Define Part, Package, Speedgrade 
###############################################################
set device       "xc7z020"
set package      "clg484"
set speed        "-1"
set part         $device$package$speed

###############################################################
###  Setup Variables
###############################################################
####flow control
set run.topSynth   0
set run.tdImpl     1

####Report and DCP controls - values: 0-required min; 1-few extra; 2-all
set verbose      1
set dcpLevel     1

####Output Directories
set synthDir  "./v7/TopDown/Synth"
set implDir   "./v7/TopDown/Implement"
set dcpDir    "./v7/TopDown/Checkpoint"

####Input Directories
set srcDir     "./Sources"
set rtlDir     "$srcDir/hdl"
set prjDir     "$srcDir/prj"
set xdcDir     "$srcDir/xdc"

####Source required Tcl Procs
source $tclDir/design_utils.tcl
source $tclDir/synth_utils.tcl
source $tclDir/impl_utils.tcl
source $tclDir/hd_floorplan_utils.tcl

###############################################################
### Top Definition
###############################################################
set top "CipherCore_Wrapper"
add_module $top
set_attribute module $top    top_level     1
set_attribute module $top    prj           $prjDir/$top.prj
set_attribute module $top    synth         ${run.topSynth}

####################################################################
### Create TopDown implementation run 
####################################################################
add_implementation TopDown1
set_attribute impl TopDown1      top          $top
set_attribute impl TopDown1      implXDC      [list $xdcDir/${top}_flpn.xdc] 
set_attribute impl TopDown1      td.impl      0
set_attribute impl TopDown1      impl         ${run.tdImpl}

add_implementation TopDown3
set_attribute impl TopDown3      top          $top
set_attribute impl TopDown3      implXDC      [list $xdcDir/${top}_flpn.xdc] 
set_attribute impl TopDown3      td.impl      0
set_attribute impl TopDown3      impl         ${run.tdImpl}
set_attribute impl TopDown3   	opt_directive         Explore
set_attribute impl TopDown3   	place_directive       Explore
set_attribute impl TopDown3   	phys_directive        Explore
set_attribute impl TopDown3   	route_directive       Explore

add_implementation TopDown4
set_attribute impl TopDown4      top          $top
set_attribute impl TopDown4      implXDC      [list $xdcDir/${top}_flpn.xdc] 
set_attribute impl TopDown4      td.impl      0
set_attribute impl TopDown4      impl         ${run.tdImpl}
set_attribute impl TopDown4   	place_directive       WLDrivenBlockPlacement
set_attribute impl TopDown4   	phys_directive        AggressiveExplore
set_attribute impl TopDown4   	route_directive       MoreGlobalIterations

add_implementation TopDown5
set_attribute impl TopDown5      top          $top
set_attribute impl TopDown5      implXDC      [list $xdcDir/${top}_flpn.xdc] 
set_attribute impl TopDown5      td.impl      0
set_attribute impl TopDown5      impl         ${run.tdImpl}
set_attribute impl TopDown5   	place_directive       ExtraPostPlacementOpt
set_attribute impl TopDown5   	route_directive       NoTimingRelaxation

add_implementation TopDown6
set_attribute impl TopDown6      top          $top
set_attribute impl TopDown6      implXDC      [list $xdcDir/${top}_flpn.xdc] 
set_attribute impl TopDown6      td.impl      0
set_attribute impl TopDown6      impl         ${run.tdImpl}
set_attribute impl TopDown6   	place_directive       WLDrivenBlockPlacement
set_attribute impl TopDown6   	phys_directive        AlternateReplication
set_attribute impl TopDown6   	route_directive       HigherDelayCost

add_implementation TopDown7
set_attribute impl TopDown7      top          $top
set_attribute impl TopDown7      implXDC      [list $xdcDir/${top}_flpn.xdc] 
set_attribute impl TopDown7      td.impl      0
set_attribute impl TopDown7      impl         ${run.tdImpl}
set_attribute impl TopDown7   	place_directive       LateBlockPlacement
set_attribute impl TopDown7   	phys_directive        AggressiveExplore
set_attribute impl TopDown7   	route_directive       HigherDelayCost

add_implementation TopDown8
set_attribute impl TopDown8      top          $top
set_attribute impl TopDown8      implXDC      [list $xdcDir/${top}_flpn.xdc] 
set_attribute impl TopDown8      td.impl      0 
set_attribute impl TopDown8      impl         ${run.tdImpl}
set_attribute impl TopDown8   	place_directive       ExtraNetDelay_high
set_attribute impl TopDown8   	route_directive       AdvancedSkewModeling

add_implementation TopDown9
set_attribute impl TopDown9      top          $top
set_attribute impl TopDown9      implXDC      [list $xdcDir/${top}_flpn.xdc] 
set_attribute impl TopDown9      td.impl      0
set_attribute impl TopDown9      impl         ${run.tdImpl}
set_attribute impl TopDown9   	place_directive       ExtraNetDelay_medium
set_attribute impl TopDown9   	route_directive       AdvancedSkewModeling

add_implementation TopDown10
set_attribute impl TopDown10      top          $top
set_attribute impl TopDown10      implXDC      [list $xdcDir/${top}_flpn.xdc] 
set_attribute impl TopDown10      td.impl      0
set_attribute impl TopDown10      impl         ${run.tdImpl}
set_attribute impl TopDown10   	place_directive       ExtraNetDelay_low
set_attribute impl TopDown10   	route_directive       AdvancedSkewModeling

add_implementation TopDown11
set_attribute impl TopDown11      top          $top
set_attribute impl TopDown11      implXDC      [list $xdcDir/${top}_flpn.xdc] 
set_attribute impl TopDown11      td.impl      0 
set_attribute impl TopDown11      impl         ${run.tdImpl}
set_attribute impl TopDown11   	place_directive       SSI_ExtraTimingOpt
set_attribute impl TopDown11   	route_directive       NoTimingRelaxation

add_implementation TopDown12
set_attribute impl TopDown12      top          $top
set_attribute impl TopDown12      implXDC      [list $xdcDir/${top}_flpn.xdc] 
set_attribute impl TopDown12      td.impl      0
set_attribute impl TopDown12      impl         ${run.tdImpl}
set_attribute impl TopDown12   	opt_directive       ExploreArea
set_attribute impl TopDown12   phys        		  0

add_implementation TopDown13
set_attribute impl TopDown13      top          $top
set_attribute impl TopDown13      implXDC      [list $xdcDir/${top}_flpn.xdc] 
set_attribute impl TopDown13      td.impl      0
set_attribute impl TopDown13      impl         ${run.tdImpl}
set_attribute impl TopDown13   	place_directive       ExtraPostPlacementOpt
set_attribute impl TopDown13   	phys_directive        AddRetime
set_attribute impl TopDown13   	route_directive       HigherDelayCost

add_implementation TopDown15
set_attribute impl TopDown15      top          $top
set_attribute impl TopDown15      implXDC      [list $xdcDir/${top}_flpn.xdc] 
set_attribute impl TopDown15      td.impl      0
set_attribute impl TopDown15      impl         ${run.tdImpl}
set_attribute impl TopDown15   	phys_directive        Explore

add_implementation TopDown16
set_attribute impl TopDown16      top          $top
set_attribute impl TopDown16      implXDC      [list $xdcDir/${top}_flpn.xdc] 
set_attribute impl TopDown16      td.impl      0
set_attribute impl TopDown16      impl         ${run.tdImpl}
set_attribute impl TopDown16   	phys_directive        Explore
set_attribute impl TopDown16   	route_options        -tns_cleanup

add_implementation TopDown17
set_attribute impl TopDown17      top          $top
set_attribute impl TopDown17      implXDC      [list $xdcDir/${top}_flpn.xdc] 
set_attribute impl TopDown17      td.impl      0
set_attribute impl TopDown17      impl         ${run.tdImpl}
set_attribute impl TopDown17   	opt_directive       RuntimeOptimized
set_attribute impl TopDown17   	place_directive     RuntimeOptimized
set_attribute impl TopDown17   	route_directive     RuntimeOptimized
set_attribute impl TopDown17   	phys       			0

add_implementation TopDown18
set_attribute impl TopDown18      top          $top
set_attribute impl TopDown18      implXDC      [list $xdcDir/${top}_flpn.xdc] 
set_attribute impl TopDown18      td.impl      0
set_attribute impl TopDown18      impl         ${run.tdImpl}
set_attribute impl TopDown18   	opt_directive       RuntimeOptimized
set_attribute impl TopDown18   	place_directive     Quick
set_attribute impl TopDown18   	route_directive     Quick
set_attribute impl TopDown18   	phys       			0

add_implementation TopDown19
set_attribute impl TopDown19      top          $top
set_attribute impl TopDown19      implXDC      [list $xdcDir/${top}_flpn.xdc] 
set_attribute impl TopDown19      td.impl      0
set_attribute impl TopDown19      impl         ${run.tdImpl}
set_attribute impl TopDown19   	place_directive     SpreadLogic_high
set_attribute impl TopDown19   	phys_directive       AlternateReplication
set_attribute impl TopDown19   	route_directive     MoreGlobalIterations

add_implementation TopDown20
set_attribute impl TopDown20      top          $top
set_attribute impl TopDown20      implXDC      [list $xdcDir/${top}_flpn.xdc] 
set_attribute impl TopDown20      td.impl      0
set_attribute impl TopDown20      impl         ${run.tdImpl}
set_attribute impl TopDown20   	place_directive     SpreadLogic_medium
set_attribute impl TopDown20   	phys_directive       AlternateReplication
set_attribute impl TopDown20   	route_directive     NoTimingRelaxation

add_implementation TopDown21
set_attribute impl TopDown21      top          $top
set_attribute impl TopDown21      implXDC      [list $xdcDir/${top}_flpn.xdc] 
set_attribute impl TopDown21      td.impl      0
set_attribute impl TopDown21      impl         ${run.tdImpl}
set_attribute impl TopDown21   	place_directive     SpreadLogic_low
set_attribute impl TopDown21   	route_directive     HigherDelayCost

add_implementation TopDown22
set_attribute impl TopDown22      top          $top
set_attribute impl TopDown22      implXDC      [list $xdcDir/${top}_flpn.xdc] 
set_attribute impl TopDown22      td.impl      0
set_attribute impl TopDown22      impl         ${run.tdImpl}

set_attribute impl TopDown22   	place_directive     SSI_SpreadSLLs
set_attribute impl TopDown22   	route_directive     NoTimingRelaxation


add_implementation TopDown23
set_attribute impl TopDown23      top          $top
set_attribute impl TopDown23      implXDC      [list $xdcDir/${top}_flpn.xdc] 
set_attribute impl TopDown23      td.impl      0
set_attribute impl TopDown23      impl         ${run.tdImpl}

set_attribute impl TopDown23   	place_directive     SSI_BalanceSLLs
set_attribute impl TopDown23   	route_directive     NoTimingRelaxation

add_implementation TopDown24
set_attribute impl TopDown24      top          $top
set_attribute impl TopDown24      implXDC      [list $xdcDir/${top}_flpn.xdc] 
set_attribute impl TopDown24      td.impl      0
set_attribute impl TopDown24      impl         ${run.tdImpl}

set_attribute impl TopDown24   	place_directive     SSI_BalanceSLRs

add_implementation TopDown25
set_attribute impl TopDown25      top          $top
set_attribute impl TopDown25      implXDC      [list $xdcDir/${top}_flpn.xdc] 
set_attribute impl TopDown25      td.impl      0
set_attribute impl TopDown25      impl         ${run.tdImpl}

set_attribute impl TopDown25   	place_directive     SSI_HighUtilSLRs

# Build the designs
source $tclDir/run.tcl

exit