#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
gen.py: Pre-arranged AETVgen usage
"""

__author__     = "Ekawat (Ice) Homsirikamol"
__copyright__  = "Copyright (c) 2015 CERG @ GMU, USA"
__license__    = "GPL 3.0"

# Core imports
import random
import sys
import AETVgen
import shutil

repo = "../../../../"
copy = False

def AES_COPA(size, testMode, verbose, decrypt, starttv):
    args = {'stdout': False, 'sizePIO': 4, 'sizeSIO': 4,
            'minSize': 0, 'maxSize': 1000,  'minSizeAD': 0, 'maxSizeAD': 1000,
            'sizeTag': 16, 'sizeKey': 16, 'sizeNpub': 12,
            'caesarLib': 'aescopav1', 'blockSize': 16, 'blockSizeAD': 16,
            'verbose': verbose, 'testMode': testMode, 'decrypt': decrypt, 'startTV': starttv,
            'plainTextMode': 2,
            'maxSizeSegment' : 10000}
    aetv = AETVgen.AETVgen(**args)
    aetv.genTestVectors(size)
    # Copy generated files to hardware repository
    if copy == True:
        dst = repo + "AES_COPA/x1/HLS/scripts/"
        for file in ["do.txt", "pdi.txt", "sdi.txt"]:
            shutil.copyfile(file, dst + file)
        dst = repo + "AES_COPA/x1/RTL/scripts/"
        for file in ["do.txt", "pdi.txt", "sdi.txt"]:
            shutil.copyfile(file, dst + file)

def AES_GCM(size, testMode, verbose, decrypt, starttv):
    args = {'stdout': False, 'sizePIO': 4, 'sizeSIO': 4,
            'minSize': 0, 'maxSize': 1000,  'minSizeAD': 0, 'maxSizeAD': 1000,
            'sizeTag': 16, 'sizeKey': 16, 'sizeNpub': 12,
            'caesarLib': 'aes128gcmv1', 'blockSize': 16, 'blockSizeAD': 16,
            'verbose': verbose, 'testMode': testMode, 'decrypt': decrypt, 'startTV': starttv,
            'maxSizeSegment' : 10000}
    aetv = AETVgen.AETVgen(**args)
    aetv.genTestVectors(size)
    # Copy generated files to hardware repository
    if copy == True:
        dst = repo + "AES_GCM/x1/HLS/scripts/"
        for file in ["do.txt", "pdi.txt", "sdi.txt"]:
            shutil.copyfile(file, dst + file)
        dst = repo + "AES_GCM/x1/RTL/scripts/"
        for file in ["do.txt", "pdi.txt", "sdi.txt"]:
            shutil.copyfile(file, dst + file)

def AES_GCM_RDKEY(size, testMode, verbose, decrypt, starttv):
    args = {'stdout': False, 'sizePIO': 4, 'sizeSIO': 4,
            'minSize': 0, 'maxSize': 1000,  'minSizeAD': 0, 'maxSizeAD': 1000,
            'sizeTag': 16, 'sizeKey': 16, 'sizeNpub': 12,
            'enableNsec': False, 'sizeNsec': 16,
            'caesarLib': 'aes128gcmv1', 'blockSize': 16, 'blockSizeAD': 16,
            'enableRoundKey': True, 'sizeRoundKey': 16, 'totalRoundKey': 11,
            'verbose': verbose, 'testMode': testMode, 'decrypt': decrypt, 'startTV': starttv,
            'maxSizeSegment' : 10000}
    aetv = AETVgen.AETVgen(**args)
    aetv.genTestVectors(size)
    # Copy generated files to hardware repository
    if copy == True:
        dst = repo + "AES_GCM/x1_rdkey/RTL/scripts/"
        for file in ["do.txt", "pdi.txt", "sdi.txt"]:
            shutil.copyfile(file, dst + file)

def ASCON(size, testMode, verbose, decrypt, starttv):
    args = {'stdout': False, 'sizePIO': 4, 'sizeSIO': 4,
            'minSize': 0, 'maxSize': 1000,  'minSizeAD': 0, 'maxSizeAD': 1000,
            'sizeTag': 16, 'sizeKey': 16, 'sizeNpub': 16,
            'caesarLib': 'ascon128v1', 'blockSize': 8, 'blockSizeAD': 8,
            'verbose': verbose, 'testMode': testMode, 'decrypt': decrypt, 'startTV': starttv,
            'maxSizeSegment' : 10000}
    aetv = AETVgen.AETVgen(**args)
    aetv.genTestVectors(size)
    # Copy generated files to hardware repository
    if copy == True:
        dst = repo + "ASCON/x1/RTL/scripts/"
        for file in ["do.txt", "pdi.txt", "sdi.txt"]:
            shutil.copyfile(file, dst + file)

def Deoxys(size, testMode, verbose, decrypt, starttv):
    args = {'stdout': False, 'sizePIO': 4, 'sizeSIO': 4,
            'minSize': 0, 'maxSize': 100, 'minSizeAD': 0, 'maxSizeAD': 100,
            'sizeTag': 16, 'sizeKey': 16, 'sizeNpub': 8,
            'caesarLib': 'deoxysneq128128v1', 'blockSize': 16, 'blockSizeAD': 16,
            'verbose': verbose, 'testMode': testMode, 'decrypt': decrypt, 'startTV': starttv,
            'maxSizeSegment' : 64}
    aetv = AETVgen.AETVgen(**args)
    aetv.genTestVectors(size)
    # Copy files to hardware repository
    if copy == True:
        dst = repo + "Deoxys/x1/RTL/scripts/"
        for file in ["do.txt", "pdi.txt", "sdi.txt"]:
            shutil.copyfile(file, dst + file)
        dst = repo + "Deoxys/x1/HLS/scripts/"
        for file in ["do.txt", "pdi.txt", "sdi.txt"]:
            shutil.copyfile(file, dst + file)

def ICEPOLE_Nsec(size, testMode, verbose, decrypt, starttv):
    args = {'stdout': False, 'sizePIO': 32, 'sizeSIO': 4, 'enableNsec': True,
            'minSize': 0, 'maxSize': 1000,  'minSizeAD': 0, 'maxSizeAD': 1000,
            'sizeTag': 16, 'sizeKey': 16, 'sizeNpub': 12, 'sizeNsec': 16,
            'caesarLib': 'icepole128v1', 'blockSize': 128, 'blockSizeAD': 128,
            'verbose': verbose, 'testMode': testMode, 'decrypt': decrypt, 'startTV': starttv,
            'maxSizeSegment' : 256}
    aetv = AETVgen.AETVgen(**args)
    aetv.genTestVectors(size)
    # Copy files to hardware repository
    if copy == True:
        dst = repo + "ICEPOLE/x1_nsec/RTL/scripts/"
        for file in ["do.txt", "pdi.txt", "sdi.txt"]:
            shutil.copyfile(file, dst + file)

def ICEPOLE(size, testMode, verbose, decrypt, starttv):
    args = {'stdout': False, 'sizePIO': 32, 'sizeSIO': 4, 'enableNsec': False,
            'minSize': 0, 'maxSize': 1000,  'minSizeAD': 0, 'maxSizeAD': 1000,
            'sizeTag': 16, 'sizeKey': 16, 'sizeNpub': 12,
            'caesarLib': 'icepole128av1', 'blockSize': 128, 'blockSizeAD': 128,
            'verbose': verbose, 'testMode': testMode, 'decrypt': decrypt, 'startTV': starttv,
            'maxSizeSegment' : 256}
    aetv = AETVgen.AETVgen(**args)
    aetv.genTestVectors(size)
    # Copy files to hardware repository
    if copy == True:
        dst = repo + "ICEPOLE/x1/HLS/scripts/"
        for file in ["do.txt", "pdi.txt", "sdi.txt"]:
            shutil.copyfile(file, dst + file)
        dst = repo + "ICEPOLE/x1/RTL/scripts/"
        for file in ["do.txt", "pdi.txt", "sdi.txt"]:
            shutil.copyfile(file, dst + file)

def Joltik(size, testMode, verbose, decrypt, starttv):
    args = {'stdout': False, 'sizePIO': 4, 'sizeSIO': 4,
            'minSize': 0, 'maxSize': 100, 'minSizeAD': 0, 'maxSizeAD': 100,
            'sizeTag': 8, 'sizeKey': 16, 'sizeNpub': 4,
            'caesarLib': 'joltikneq12864v1', 'blockSize': 8, 'blockSizeAD': 8,
            'verbose': verbose, 'testMode': testMode, 'decrypt': decrypt, 'startTV': starttv,
            'maxSizeSegment' : 32}
    aetv = AETVgen.AETVgen(**args)
    aetv.genTestVectors(size)
    # Copy files to hardware repository
    if copy == True:
        dst = repo + "Joltik/x1/RTL/scripts/"
        for file in ["do.txt", "pdi.txt", "sdi.txt"]:
            shutil.copyfile(file, dst + file)
        dst = repo + "Joltik/x1/HLS/scripts/"
        for file in ["do.txt", "pdi.txt", "sdi.txt"]:
            shutil.copyfile(file, dst + file)

def Ketje(size, testMode, verbose, decrypt, starttv):
    args = {'stdout': False, 'sizePIO': 16, 'sizeSIO': 4,
            'minSize': 0, 'maxSize': 1000, 'minSizeAD': 0, 'maxSizeAD': 1000,
            'sizeTag': 8, 'sizeKey': 16, 'sizeNpub': 16,
            'caesarLib': 'ketjejrv1', 'blockSize': 200/8, 'blockSizeAD': 200/8,
            'verbose': verbose, 'testMode': testMode, 'decrypt': decrypt, 'startTV': starttv,
            'maxSizeSegment' : 64}
    aetv = AETVgen.AETVgen(**args)
    aetv.genTestVectors(size)
    # Copy files to hardware repository
    if copy == True:
        dst = repo + "Ketje/x1/RTL/scripts/"
        for file in ["do.txt", "pdi.txt", "sdi.txt"]:
            shutil.copyfile(file, dst + file)
        dst = repo + "Ketje/x1/HLS/scripts/"
        for file in ["do.txt", "pdi.txt", "sdi.txt"]:
            shutil.copyfile(file, dst + file)

def Keyak(size, testMode, verbose, decrypt, starttv):
    args = {'stdout': False, 'sizePIO': 16, 'sizeSIO': 4,
            'minSize': 0, 'maxSize': 1000, 'minSizeAD': 0, 'maxSizeAD': 1000,
            'sizeTag': 16, 'sizeKey': 16, 'sizeNpub': 16,
            'caesarLib': 'lakekeyakv1', 'blockSize': 1344/8, 'blockSizeAD': 1344/8,
            'verbose': verbose, 'testMode': testMode, 'decrypt': decrypt, 'startTV': starttv,
            'plainTextMode': 1,
            'maxSizeSegment' : 1280}
    aetv = AETVgen.AETVgen(**args)
    aetv.genTestVectors(size)
    # Copy files to hardware repository
    if copy == True:
        dst = repo + "Keyak/LakeKeyak/x1/RTL/scripts/"
        for file in ["do.txt", "pdi.txt", "sdi.txt"]:
            shutil.copyfile(file, dst + file)
        dst = repo + "Keyak/LakeKeyak/x1/HLS/scripts/"
        for file in ["do.txt", "pdi.txt", "sdi.txt"]:
            shutil.copyfile(file, dst + file)


def Minalpher(size, testMode, verbose, decrypt, starttv):
    args = {'stdout': False, 'sizePIO': 4, 'sizeSIO': 4,
            'sizeTag': 16, 'sizeKey': 16, 'sizeNpub': 13,
            'minSize': 0, 'maxSize': 1000,
            'minSizeAD': 0, 'maxSizeAD': 1000, 'cipherTextMode' : 2,
            'caesarLib': 'minalpherv1', 'blockSize': 256/8, 'blockSizeAD': 256/8,
            'cipherTextMode': 2, 'padD': 4, 'padAD': 1,
            'verbose': verbose, 'testMode': testMode, 'decrypt': decrypt, 'startTV': starttv,
            'maxSizeSegment' : 64,
            'enableLoadLen': False, 'ctrSizeAd': 2, 'ctrSizeD': 5,
            'genCipherCoreTvC': True, 'Latency': {'EncTagCycles': 1, 'DecTagCycles': 0, 'NoMsgTagCycles': 0}}
    aetv = AETVgen.AETVgen(**args)
    aetv.genTestVectors(size)
    # Copy files to hardware repository
    if copy == True:
        dst = repo + "Minalpher/x1/HLS/scripts/"
        for file in ["do.txt", "pdi.txt", "sdi.txt"]:
            shutil.copyfile(file, dst + file)
        dst = repo + "Minalpher/x1/RTL/scripts/"
        for file in ["do.txt", "pdi.txt", "sdi.txt"]:
            shutil.copyfile(file, dst + file)

def OCB(size, testMode, verbose, decrypt, starttv):
    args = {'stdout': False, 'sizePIO': 4, 'sizeSIO': 4,
            'sizeTag': 16, 'sizeKey': 16, 'sizeNpub': 12,
            'minSize': 0, 'maxSize': 1000,
            'minSizeAD': 0, 'maxSizeAD': 1000,
            'caesarLib': 'aeadaes128ocbtaglen128v1', 'blockSize': 16, 'blockSizeAD': 16,
            'verbose': verbose, 'testMode': testMode, 'decrypt': decrypt, 'startTV': starttv,
            'maxSizeSegment' : 800}
    aetv = AETVgen.AETVgen(**args)
    aetv.genTestVectors(size)
    # Copy files to hardware repository
    if copy == True:
        dst = repo + "OCB/x1/HLS/scripts/"
        for file in ["do.txt", "pdi.txt", "sdi.txt"]:
            shutil.copyfile(file, dst + file)
        dst = repo + "OCB/x1/RTL/scripts/"
        for file in ["do.txt", "pdi.txt", "sdi.txt"]:
            shutil.copyfile(file, dst + file)


def PAEQ(size, testMode, verbose, decrypt, starttv):
    args = {'stdout': False, 'sizePIO': 4, 'sizeSIO': 4,
            'sizeTag': 16, 'sizeKey': 16, 'sizeNpub': 12,
            'minSize': 0, 'maxSize': 1000,
            'minSizeAD': 0, 'maxSizeAD': 1000,
            'caesarLib': 'paeq128', 'blockSize': 46, 'blockSizeAD': 30,
            'verbose': verbose, 'testMode': testMode, 'decrypt': decrypt, 'startTV': starttv,
            'maxSizeSegment' : 800}
    aetv = AETVgen.AETVgen(**args)
    aetv.genTestVectors(size)
    # Copy files to hardware repository
    if copy == True:
        dst = repo + "PAEQ/x1/HLS/scripts/"
        for file in ["do.txt", "pdi.txt", "sdi.txt"]:
            shutil.copyfile(file, dst + file)
        dst = repo + "PAEQ/x1/RTL/scripts/"
        for file in ["do.txt", "pdi.txt", "sdi.txt"]:
            shutil.copyfile(file, dst + file)

def POET(size, testMode, verbose, decrypt, starttv):
    args = {'stdout': False, 'sizePIO': 4, 'sizeSIO': 4,
            'sizeTag': 16, 'sizeKey': 16, 'sizeNpub': 12,
            'minSize': 0, 'maxSize': 1000,
            'minSizeAD': 0, 'maxSizeAD': 1000,
            'caesarLib': 'aes128poetv1aes128', 'blockSize': 16, 'blockSizeAD': 16,
            'verbose': verbose, 'testMode': testMode, 'decrypt': decrypt, 'startTV': starttv,
            'maxSizeSegment' : 800}
    aetv = AETVgen.AETVgen(**args)
    aetv.genTestVectors(size)
    # Copy files to hardware repository
    if copy == True:
        dst = repo + "POET/x1/HLS/scripts/"
        for file in ["do.txt", "pdi.txt", "sdi.txt"]:
            shutil.copyfile(file, dst + file)
        dst = repo + "POET/x1/RTL/scripts/"
        for file in ["do.txt", "pdi.txt", "sdi.txt"]:
            shutil.copyfile(file, dst + file)


def PRIMATEs_APE(size, testMode, verbose, decrypt, starttv):
    args = {'stdout': False, 'sizePIO': 5, 'sizeSIO': 5, 'enableNsec': False,
            'sizeTag': 30, 'sizeKey': 30, 'sizeNpub': 15,
            'minSize': 0, 'maxSize': 300, 'minSizeAD': 0, 'maxSizeAD': 128,
            'caesarLib': 'primatesv1ape120', 'blockSize': 5, 'blockSizeAD': 5,
            'plainTextMode': 1, 'cipherTextMode': 2, 'reverseDblk': True,
            'verbose': verbose, 'testMode': testMode, 'decrypt': decrypt, 'startTV': starttv,
            'maxSizeSegment' : 10000, 'padStyle': 0,
            'genCipherCoreTvC': True, 'Latency': {'EncTagCycles': 0, 'DecTagCycles': 0, 'NoMsgTagCycles': 1}}
    aetv = AETVgen.AETVgen(**args)
    aetv.genTestVectors(size)
    # Copy files to hardware repository
    if copy == True:
        dst = repo + "PRIMATEs/APE/x1/HLS/scripts/"
        for file in ["do.txt", "pdi.txt", "sdi.txt"]:
            shutil.copyfile(file, dst + file)
        dst = repo + "PRIMATEs/APE/x1/RTL/scripts/"
        for file in ["do.txt", "pdi.txt", "sdi.txt"]:
            shutil.copyfile(file, dst + file)

def PRIMATEs_GIBBON(size, testMode, verbose, decrypt, starttv):
    args = {'stdout': False, 'sizePIO': 5, 'sizeSIO': 5, 'enableNsec': False,
            'sizeTag': 15, 'sizeKey': 15, 'sizeNpub': 15,
            'minSize': 0, 'maxSize': 300, 'minSizeAD': 0, 'maxSizeAD': 150,
            'caesarLib': 'primatesv1gibbon120', 'blockSize': 5, 'blockSizeAD': 5,
            'verbose': verbose, 'testMode': testMode, 'decrypt': decrypt, 'startTV': starttv,
            'maxSizeSegment' : 10000}
    aetv = AETVgen.AETVgen(**args)
    aetv.genTestVectors(size)
    # Copy files to hardware repository
    if copy == True:
        dst = repo + "PRIMATEs/GIBBON/x1/HLS/scripts/"
        for file in ["do.txt", "pdi.txt", "sdi.txt"]:
            shutil.copyfile(file, dst + file)
        dst = repo + "PRIMATEs/GIBBON/x1/RTL/scripts/"
        for file in ["do.txt", "pdi.txt", "sdi.txt"]:
            shutil.copyfile(file, dst + file)

def PRIMATEs_HANUMAN(size, testMode, verbose, decrypt, starttv):
    args = {'stdout': False, 'sizePIO': 5, 'sizeSIO': 5, 'enableNsec': False,
            'sizeTag': 15, 'sizeKey': 15, 'sizeNpub': 15,
            'minSize': 0, 'maxSize': 300, 'minSizeAD': 0, 'maxSizeAD': 150,
            'caesarLib': 'primatesv1hanuman120', 'blockSize': 5, 'blockSizeAD': 5,
            'verbose': verbose, 'testMode': testMode, 'decrypt': decrypt, 'startTV': starttv,
            'maxSizeSegment' : 10000}
    aetv = AETVgen.AETVgen(**args)
    aetv.genTestVectors(size)
    # Copy files to hardware repository
    if copy == True:
        dst = repo + "PRIMATEs/HANUMAN/x1/HLS/scripts/"
        for file in ["do.txt", "pdi.txt", "sdi.txt"]:
            shutil.copyfile(file, dst + file)
        dst = repo + "PRIMATEs/HANUMAN/x1/RTL/scripts/"
        for file in ["do.txt", "pdi.txt", "sdi.txt"]:
            shutil.copyfile(file, dst + file)

def Scream(size, testMode, verbose, decrypt, starttv):
    args = {'stdout': False, 'sizePIO': 4, 'sizeSIO': 4,
            'minSize': 0, 'maxSize': 32, 'minSizeAD': 0, 'maxSizeAD': 32,
            'sizeTag': 16, 'sizeKey': 16, 'sizeNpub': 12,
            'caesarLib': 'Scream10v2', 'blockSize': 16, 'blockSizeAD': 16,
            'verbose': verbose, 'testMode': testMode, 'decrypt': decrypt, 'startTV': starttv,
            'maxSizeSegment' : 10000}
    aetv = AETVgen.AETVgen(**args)
    aetv.genTestVectors(size)
    # Copy files to hardware repository
    if copy == True:
        dst = repo + "Joltik/x1/RTL/scripts/"
        for file in ["do.txt", "pdi.txt", "sdi.txt"]:
            shutil.copyfile(file, dst + file)

if __name__ == '__main__':
    testMode  = 1
    startTV   = 0
    totalTest = 20
    verbose   = False
    #verbose   = False
    decrypt   = False
    #AES_COPA(totalTest, testMode, verbose, decrypt, startTV)
    AES_GCM(totalTest, testMode, verbose, decrypt, startTV)
    #AES_GCM_RDKEY(totalTest, testMode, verbose, decrypt, startTV)
    #ASCON(totalTest, testMode, verbose, decrypt, startTV)
    #Deoxys(totalTest, testMode, verbose, decrypt, startTV)
    #ICEPOLE(totalTest, testMode, verbose, decrypt, startTV)
    #ICEPOLE_Nsec(totalTest, testMode, verbose, decrypt, startTV)
    #Joltik(totalTest, testMode, verbose, decrypt, startTV)
    #Keyak(totalTest, testMode, verbose, decrypt, startTV)
    #Minalpher(totalTest, testMode, verbose, decrypt, startTV)
    #OCB(totalTest, testMode, verbose, decrypt, startTV)
    #PAEQ(totalTest, testMode, verbose, decrypt, startTV)
    #PRIMATEs_APE(totalTest, testMode, verbose, decrypt, startTV)
    #PRIMATEs_GIBBON(totalTest, testMode, verbose, decrypt, startTV)
    #PRIMATEs_HANUMAN(totalTest, testMode, verbose, decrypt, startTV)
    #Scream(totalTest, testMode, verbose, decrypt, startTV)
